@@ -20,7 +20,6 @@
 #include "str.h"
 #include "set.h"
 
-#define INDEX_SIZE (256U * sizeof(void *))
 #define INITIAL_N_ENTRIES 256U
 #define INITIAL_INDEX_SIZE 256U
 
@@ -104,8 +103,8 @@ sen_set_reset(sen_set * set, uint32_t ne)
       e = *sp; 
       if (!e || (e == GARBAGE)) { continue; } 
       for (i = e->key; ; i += 1048573) { 
-	dp = index + (i & m); 
-	if (!*dp) { break; } 
+        dp = index + (i & m); 
+        if (!*dp) { break; } 
       } 
       *dp = e; 
     } 
@@ -114,8 +113,8 @@ sen_set_reset(sen_set * set, uint32_t ne)
       e = *sp; 
       if (!e || (e == GARBAGE)) { continue; } 
       for (i = ((entry_str *)e)->key; ; i += 1048573) { 
-	dp = index + (i & m); 
-	if (!*dp) { break; } 
+        dp = index + (i & m); 
+        if (!*dp) { break; } 
       } 
       *dp = e; 
     } 
@@ -153,7 +152,7 @@ sen_set_close(sen_set * set)
 
 sen_rc
 sen_set_info(sen_set *set, unsigned *key_size, unsigned *value_size,
-	     unsigned *n_entries)
+             unsigned *n_entries)
 {
   if (!set) { return sen_invalid_argument; }
   if (key_size) { *key_size = set->key_size; }
@@ -303,7 +302,7 @@ sen_set_bin_get(sen_set *set, const void *key, void **value)
       if (!np) { np = ep; }
     } else {
       if (e->key == h && !memcmp(key, e->dummy, set->key_size)) {
-	goto exit;
+        goto exit;
       }
     }
   }
@@ -501,8 +500,8 @@ sort_int(sen_set *set, entry **res, int limit, int offset, int dir)
 
 inline static entry **
 pack_func(sen_set *set, entry **res,
-	  int(*func)(sen_set *, entry **, sen_set *, entry **, void *),
-	  void *arg, void *arg0, int dir)
+          int(*func)(sen_set *, entry **, sen_set *, entry **, void *),
+          void *arg, void *arg0, int dir)
 {
   uint32_t i, n, m = set->max_offset;
   entry **head, **tail, *e, *c;
@@ -529,8 +528,8 @@ pack_func(sen_set *set, entry **res,
 
 inline static entry **
 part_func(entry **b, entry **e,
-	  int(*func)(sen_set *, entry **, sen_set *, entry **, void *),
-	  void *arg, void *arg0, int dir)
+          int(*func)(sen_set *, entry **, sen_set *, entry **, void *),
+          void *arg, void *arg0, int dir)
 {
   intptr_t d = e - b;
   entry **c;
@@ -558,8 +557,8 @@ part_func(entry **b, entry **e,
 
 static void
 _sort_func(entry **head, entry **tail, int limit,
-	   int(*func)(sen_set *, entry **, sen_set *, entry **, void *),
-	   void *arg, void *arg0, int dir)
+           int(*func)(sen_set *, entry **, sen_set *, entry **, void *),
+           void *arg, void *arg0, int dir)
 {
   entry **c;
   if (head < tail && (c = part_func(head, tail, func, arg, arg0, dir))) {
@@ -571,8 +570,8 @@ _sort_func(entry **head, entry **tail, int limit,
 
 inline static void
 sort_func(sen_set *set, entry **res, int limit,
-	  int(*func)(sen_set *, entry **, sen_set *, entry **, void *),
-	  void *arg, void *arg0, int dir)
+          int(*func)(sen_set *, entry **, sen_set *, entry **, void *),
+          void *arg, void *arg0, int dir)
 {
   entry **c = pack_func(set, res, func, arg, arg0, dir);
   if (c) {
@@ -601,8 +600,16 @@ sen_set_sort(sen_set *set, int limit, sen_set_sort_optarg *optarg)
 {
   entry **res;
   int dir = 1;
-  if (!set->n_entries) { return NULL; }
+  if (!set) {
+    SEN_LOG(sen_log_warning, "sen_set_sort: invalid argument !");
+    return NULL;
+  }
+  if (!set->n_entries) {
+    SEN_LOG(sen_log_warning, "no entry in the set passed for sen_set_sort");
+    return NULL;
+  }
   if (!(res = SEN_MALLOC(sizeof(entry *) * set->n_entries))) {
+    SEN_LOG(sen_log_alert, "allocation of entries failed on sen_set_sort !");
     return NULL;
   }
   if (!limit || limit > set->n_entries) { limit = set->n_entries; }
@@ -610,7 +617,7 @@ sen_set_sort(sen_set *set, int limit, sen_set_sort_optarg *optarg)
     dir = (optarg->mode == sen_sort_ascending) ? 1 : -1;
     if (optarg->compar) {
       sort_func(set, res, limit, optarg->compar, optarg->compar_arg,
-		optarg->compar_arg0 ? optarg->compar_arg0 : set, dir);
+                optarg->compar_arg0 ? optarg->compar_arg0 : set, dir);
       goto exit;
     } else if (optarg->compar_arg) {
       sort_int(set, res, limit, ((intptr_t)optarg->compar_arg) / sizeof(int32_t), dir);
@@ -664,23 +671,23 @@ sen_set_union(sen_set *a, sen_set *b)
     if ((e = *ep) && e != GARBAGE) {
       switch (key_size) {
       case 0 :
-	key = SEN_SET_STRKEY(e);
-	vb = SEN_SET_STRVAL(e);
-	break;
+        key = SEN_SET_STRKEY(e);
+        vb = SEN_SET_STRVAL(e);
+        break;
       case sizeof(uint32_t) :
-	key = SEN_SET_INTKEY(e);
-	vb = SEN_SET_INTVAL(e);
-	break;
+        key = SEN_SET_INTKEY(e);
+        vb = SEN_SET_INTVAL(e);
+        break;
       default :
-	key = SEN_SET_BINKEY(e);
-	vb = &e->dummy[key_size];
-	break;
+        key = SEN_SET_BINKEY(e);
+        vb = &e->dummy[key_size];
+        break;
       }
       if (sen_set_at(a, key, &va)) {
-	/* do copy of merge? */
+        /* do copy of merge? */
       } else {
-	sen_set_get(a, key, &va);
-	memcpy(va, vb, value_size);
+        sen_set_get(a, key, &va);
+        memcpy(va, vb, value_size);
       }
       i--;
     }
@@ -700,14 +707,14 @@ sen_set_subtract(sen_set *a, sen_set *b)
     if ((e = *ep) && e != GARBAGE) {
       switch (key_size) {
       case 0 :
-	key = SEN_SET_STRKEY(e);
-	break;
+        key = SEN_SET_STRKEY(e);
+        break;
       case sizeof(uint32_t) :
-	key = SEN_SET_INTKEY(e);
-	break;
+        key = SEN_SET_INTKEY(e);
+        break;
       default :
-	key = SEN_SET_BINKEY(e);
-	break;
+        key = SEN_SET_BINKEY(e);
+        break;
       }
       if ((dp = sen_set_at(a, key, NULL))) { sen_set_del(a, dp); }
       i--;
@@ -728,14 +735,14 @@ sen_set_intersect(sen_set *a, sen_set *b)
     if ((e = *ep) && e != GARBAGE) {
       switch (key_size) {
       case 0 :
-	key = SEN_SET_STRKEY(e);
-	break;
+        key = SEN_SET_STRKEY(e);
+        break;
       case sizeof(uint32_t) :
-	key = SEN_SET_INTKEY(e);
-	break;
+        key = SEN_SET_INTKEY(e);
+        break;
       default :
-	key = SEN_SET_BINKEY(e);
-	break;
+        key = SEN_SET_BINKEY(e);
+        break;
       }
       if (!sen_set_at(b, key, NULL)) { sen_set_del(a, ep); }
       i--;
@@ -756,19 +763,19 @@ sen_set_difference(sen_set *a, sen_set *b)
     if ((e = *ep) && e != GARBAGE) {
       switch (key_size) {
       case 0 :
-	key = SEN_SET_STRKEY(e);
-	break;
+        key = SEN_SET_STRKEY(e);
+        break;
       case sizeof(uint32_t) :
-	key = SEN_SET_INTKEY(e);
-	break;
+        key = SEN_SET_INTKEY(e);
+        break;
       default :
-	key = SEN_SET_BINKEY(e);
-	break;
+        key = SEN_SET_BINKEY(e);
+        break;
       }
       if ((dp = sen_set_at(b, key, NULL))) {
-	sen_set_del(b, dp);
-	sen_set_del(a, ep);
-	count++;
+        sen_set_del(b, dp);
+        sen_set_del(a, ep);
+        count++;
       }
       i--;
     }
@@ -779,7 +786,9 @@ sen_set_difference(sen_set *a, sen_set *b)
 sen_rc
 sen_set_array_init(sen_set *set, uint32_t size)
 {
+  SEN_ASSERT(!set->n_entries);
   set->arrayp = 1;
+  set->curr_entry = 0;
   if (set->chunks[SEN_SET_MAX_CHUNK]) {
     SEN_FREE(set->chunks[SEN_SET_MAX_CHUNK]);
   }
