@@ -1,27 +1,9 @@
 /* 
- * Copyright (c) 2006 NTT Data corporation.
+ * Copyright (c) 2006-2007 NTT DATA CORPORATION.
  * All rights reserved.
  *
  * derived from original pg_senna.c developed by Daisuke Maki
  * (See below for copyright notice)
- * 
- * Permission to use, copy, modify, and distribute this software and
- * its documentation for any purpose, without fee, and without a
- * written agreement is hereby granted, provided that the above
- * copyright notice and this paragraph and the following two
- * paragraphs appear in all copies.
- *
- * IN NO EVENT SHALL THE AUTHOR BE LIABLE TO ANY PARTY FOR DIRECT,
- * INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
- * LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
- * DOCUMENTATION, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED
- * OF THE POSSIBILITY OF SUCH DAMAGE.
- *
- * THE AUTHOR SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
- * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
- * A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS
- * IS" BASIS, AND THE AUTHOR HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE,
- * SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
  *
  */
 /* 
@@ -99,6 +81,14 @@
 #include "optimizer/restrictinfo.h"
 #include "nodes/relation.h"
 #include "storage/itemptr.h"
+#include "utils/guc.h"
+#include "funcapi.h"
+
+#ifdef POSTGRES82
+#include "access/reloptions.h"
+
+PG_MODULE_MAGIC;
+#endif
 
 PG_FUNCTION_INFO_V1(pgs2insert);
 PG_FUNCTION_INFO_V1(pgs2beginscan);
@@ -110,22 +100,29 @@ PG_FUNCTION_INFO_V1(pgs2markpos);
 PG_FUNCTION_INFO_V1(pgs2restrpos);
 PG_FUNCTION_INFO_V1(pgs2build);
 PG_FUNCTION_INFO_V1(pgs2buildb);
+PG_FUNCTION_INFO_V1(pgs2buildu);
 PG_FUNCTION_INFO_V1(pgs2bulkdelete);
+PG_FUNCTION_INFO_V1(pgs2vacuumcleanup);
 PG_FUNCTION_INFO_V1(pgs2costestimate);
 PG_FUNCTION_INFO_V1(pgs2contain);
 PG_FUNCTION_INFO_V1(pgs2nop);
 PG_FUNCTION_INFO_V1(pgs2getscore);
 PG_FUNCTION_INFO_V1(pgs2getscore_simple);
 PG_FUNCTION_INFO_V1(pgs2getnhits);
-PG_FUNCTION_INFO_V1(pgs2destroy);
+#ifdef POSTGRES82
+PG_FUNCTION_INFO_V1(pgs2options);
+#endif
+PG_FUNCTION_INFO_V1(pgs2indexcache);
 
-typedef struct {
+typedef struct _index_info{
   Oid relid;
   Oid namespace;
   char name[NAMEDATALEN];
   sen_index *index;
   int using;
   IndexScanDesc s;
+  struct _index_info *previous;
+  struct _index_info *next;
 } index_info;
 
 typedef struct {
@@ -144,19 +141,11 @@ typedef struct {
   sen_recordh *mark;
 } scan_stat;
 
-static index_info index_cache[PGS2_MAX_N_INDEX_CACHE];
+static index_info *index_cache;
+static index_info *last_used_cache;
 static scan_stat *curr_scan_stat;
 static int last_nhits = 0;
-
-inline static char *
-text2cstr(text *t)
-{
-  int32 size = VARSIZE(t) - VARHDRSZ;
-  char *c = palloc(size + 1);
-  memcpy(c, t->vl_dat, size);
-  c[size] = '\0';
-  return c;
-}
+static int max_n_index_cache = 16;
 
 inline static void
 sen_check_init(void)
@@ -165,18 +154,53 @@ sen_check_init(void)
   int i;
   sen_rc rc;
   index_info *ii;
+  const char *option;
+
   if (initialized) { return ; }
   if ((rc = sen_init())) { elog(ERROR, "pgsenna2: sen_init() failed %d", rc); }
-  for (i = PGS2_MAX_N_INDEX_CACHE, ii = index_cache; i; i--, ii++) {
+  option = GetConfigOption("ludia.max_n_index_cache");
+  if (option) {
+    max_n_index_cache = atoi(option);
+    if(max_n_index_cache <= 0) {
+      elog(ERROR, "pgsenna2: value of max_n_index_cache is invalid: %d", max_n_index_cache);
+    }
+  } else {
+    elog(WARNING, "pgsenna2: value of max_n_index_cache = %d", max_n_index_cache);
+  }
+  index_cache = (index_info*)malloc(max_n_index_cache * sizeof(*index_cache));
+  last_used_cache = NULL;
+  for (i = max_n_index_cache, ii = index_cache; i; i--, ii++) {
     ii->namespace = 0;
     ii->relid = 0;
     ii->name[0] = '\0';
     ii->index = NULL;
+    ii->using = 0;
     ii->s = NULL;
+    ii->previous = NULL;
+    ii->next = NULL;
   }
   initialized++;
 }
 
+inline static void
+index_list_update(index_info *ii) {
+  if (last_used_cache == ii) {
+    return;
+  }
+  if (ii->previous) {
+    ((index_info *)ii->previous)->next = ii->next;
+  }
+  if (ii->next) {
+    ((index_info *)ii->next)->previous = ii->previous;
+  }
+  if (last_used_cache) {
+    last_used_cache->previous = ii;
+  }
+  ii->next = last_used_cache;
+  ii->previous = NULL;
+  last_used_cache = ii;
+}
+
 /* todo: SizeOfIptrData may be more suitable.  */
 #define KEY_SIZE (sizeof(ItemPointerData))
 
@@ -190,19 +214,40 @@ index_info_open(Relation r, int createp, int flags)
   Oid relid = RelationGetRelid(r);
   char *name = RelationGetRelationName(r);
   sen_check_init();
-  cand = NULL;
-  for (i = PGS2_MAX_N_INDEX_CACHE, ii = index_cache; i; i--, ii++) {
-    if (ii->index && 
-        namespace == ii->namespace &&
-        relid == ii->relid && 
-        !strcmp(ii->name, name)) {
-      return ii;
+  ii = last_used_cache;
+  for (i = max_n_index_cache - 1; i; i--) {
+    if(!ii) {
+      break;
+    } else {
+      if (ii->index && 
+          namespace == ii->namespace &&
+          relid == ii->relid && 
+          !strcmp(ii->name, name)) {
+        if (createp) { /* for REINDEX (and CLUSTER) */
+          break;
+        } else {
+          index_list_update(ii);
+          return ii;
+        }
+      } else {
+        ii = ii->next;
+      }
+    }
+  }
+  cand = ii;  // cand is oldest or NULL
+  if (!cand) {
+    for (i = max_n_index_cache, ii = index_cache; i; i--, ii++) {
+      if (!ii->using) {
+        cand = ii;
+        break;
+      }
     }
-    if (!ii->using) { cand = ii; }
   }
   if (!cand) {
-    elog(ERROR, "pgsenna2: n of indices in use reached max(%d)", PGS2_MAX_N_INDEX_CACHE);
+    elog(ERROR, "pgsenna2: n of indices in use reached max(%d)", max_n_index_cache);
   }
+  index_list_update(cand);
+
   sen_index_close(cand->index);
   cand->namespace = namespace;
   cand->relid = relid;
@@ -230,6 +275,7 @@ index_info_open(Relation r, int createp, int flags)
     default:
       encoding = sen_enc_default;
     }
+    sen_index_remove(path);
     cand->index = sen_index_create(path,
                                    KEY_SIZE,
                                    flags, 0, encoding);
@@ -299,7 +345,7 @@ do_insert(Relation r, ItemPointer tid, index_info *ii,
       t = DatumGetTextP(value);
       c = text2cstr(t);
       if (c && *c) {
-        rc = sen_index_upd(ii->index, (void *)tid, NULL, c);
+        rc = sen_index_upd(ii->index, (void *)tid, NULL, 0, c, strlen(c));
       } else {
         if (!(sen_sym_get(ii->index->keys, (void *)tid))) { rc = sen_other_error; }
       }
@@ -338,6 +384,9 @@ pgs2build(PG_FUNCTION_ARGS)
   Relation  heap = (Relation) PG_GETARG_POINTER(0);
   Relation  index = (Relation) PG_GETARG_POINTER(1);
   IndexInfo  *indexInfo = (IndexInfo *) PG_GETARG_POINTER(2);
+#ifdef POSTGRES82
+  IndexBuildResult *result;
+#endif
   build_stat bs;
   double    reltuples;
   init_support_funcs(&bs.funcs, index);
@@ -345,9 +394,17 @@ pgs2build(PG_FUNCTION_ARGS)
   bs.index = index_info_open(index, 1, SEN_INDEX_NORMALIZE);
   reltuples = IndexBuildHeapScan(heap, index, indexInfo,
                                  buildCallback, (void *) &bs);
+#ifdef POSTGRES82
+  result = (IndexBuildResult *) palloc(sizeof(IndexBuildResult));
+  result->heap_tuples = reltuples;
+  result->index_tuples = bs.indtuples;
+  index_info_close(bs.index);
+  PG_RETURN_POINTER(result);
+#else
   IndexCloseAndUpdateStats(heap, reltuples, index, bs.indtuples);
   index_info_close(bs.index);
   PG_RETURN_VOID();
+#endif
 }
 
 Datum
@@ -356,6 +413,9 @@ pgs2buildb(PG_FUNCTION_ARGS)
   Relation  heap = (Relation) PG_GETARG_POINTER(0);
   Relation  index = (Relation) PG_GETARG_POINTER(1);
   IndexInfo  *indexInfo = (IndexInfo *) PG_GETARG_POINTER(2);
+#ifdef POSTGRES82
+  IndexBuildResult *result;
+#endif
   build_stat bs;
   double    reltuples;
   init_support_funcs(&bs.funcs, index);
@@ -363,9 +423,57 @@ pgs2buildb(PG_FUNCTION_ARGS)
   bs.index = index_info_open(index, 1, SEN_INDEX_NGRAM|SEN_INDEX_NORMALIZE);
   reltuples = IndexBuildHeapScan(heap, index, indexInfo,
                                  buildCallback, (void *) &bs);
+#ifdef POSTGRES82
+  result = (IndexBuildResult *) palloc(sizeof(IndexBuildResult));
+  result->heap_tuples = reltuples;
+  result->index_tuples = bs.indtuples;
+  index_info_close(bs.index);
+  PG_RETURN_POINTER(result);
+#else
   IndexCloseAndUpdateStats(heap, reltuples, index, bs.indtuples);
   index_info_close(bs.index);
   PG_RETURN_VOID();
+#endif
+}
+
+Datum
+pgs2buildu(PG_FUNCTION_ARGS)
+{
+  Relation  heap = (Relation) PG_GETARG_POINTER(0);
+  Relation  index = (Relation) PG_GETARG_POINTER(1);
+  IndexInfo  *indexInfo = (IndexInfo *) PG_GETARG_POINTER(2);
+#ifdef POSTGRES82
+  IndexBuildResult *result;
+#endif
+  build_stat bs;
+  double    reltuples;
+  init_support_funcs(&bs.funcs, index);
+  bs.indtuples = 0;
+  const char *option;
+  int flags;
+
+  option = GetConfigOption("ludia.sen_index_flags");
+  if (option == NULL) {
+    elog(ERROR, "pgsenna2: value of sen_index_flags is invalid: NULL.");
+  }
+  flags = atoi(option);
+  if (flags < 0) {
+    elog(ERROR, "pgsenna2: value of sen_index_flags is invalid: %d.", flags);
+  }
+  bs.index = index_info_open(index, 1, flags);
+  reltuples = IndexBuildHeapScan(heap, index, indexInfo,
+                                 buildCallback, (void *) &bs);
+#ifdef POSTGRES82
+  result = (IndexBuildResult *) palloc(sizeof(IndexBuildResult));
+  result->heap_tuples = reltuples;
+  result->index_tuples = bs.indtuples;
+  index_info_close(bs.index);
+  PG_RETURN_POINTER(result);
+#else
+  IndexCloseAndUpdateStats(heap, reltuples, index, bs.indtuples);
+  index_info_close(bs.index);
+  PG_RETURN_VOID();
+#endif
 }
 
 Datum
@@ -392,9 +500,17 @@ pgs2insert(PG_FUNCTION_ARGS)
 Datum
 pgs2bulkdelete(PG_FUNCTION_ARGS)
 {
+#ifdef POSTGRES82
+  IndexVacuumInfo *info = (IndexVacuumInfo *) PG_GETARG_POINTER(0);
+  //IndexBulkDeleteResult *stats = (IndexBulkDeleteResult *) PG_GETARG_POINTER(1);
+  IndexBulkDeleteCallback callback = (IndexBulkDeleteCallback) PG_GETARG_POINTER(2);
+  void *callback_state = (void *) PG_GETARG_POINTER(3);
+  Relation r = info->index;
+#else
   Relation r = (Relation) PG_GETARG_POINTER(0);
   IndexBulkDeleteCallback callback = (IndexBulkDeleteCallback) PG_GETARG_POINTER(1);
   void *callback_state = (void *) PG_GETARG_POINTER(2);
+#endif
   double tuples_removed = 0;
   IndexBulkDeleteResult *result;
   index_info *ii = index_info_open(r, 0, 0);
@@ -411,17 +527,32 @@ pgs2bulkdelete(PG_FUNCTION_ARGS)
       if (callback(&tid, callback_state)) {
         rc = sen_index_del(ii->index, &tid);
         if (rc) { sen_log("sen_index_del failed(%d)", rc); }
+    /*
+        rc = sen_sym_del(ii->index->keys, &tid);
+        if (rc) { sen_log("sen_sym_del failed(%d)", rc); }
+    */
         tuples_removed += 1;
       }
     }
   }
   result = (IndexBulkDeleteResult *) palloc0(sizeof(IndexBulkDeleteResult));
   result->num_pages = 1;
-  result->num_index_tuples = nrecords;
+  result->num_index_tuples = sen_sym_size(ii->index->keys);
   result->tuples_removed = tuples_removed;
   PG_RETURN_POINTER(result);
 }
 
+Datum
+pgs2vacuumcleanup(PG_FUNCTION_ARGS)
+{
+  //IndexVacuumInfo *info = (IndexVacuumInfo *) PG_GETARG_POINTER(0);
+  IndexBulkDeleteResult *stats = (IndexBulkDeleteResult *) PG_GETARG_POINTER(1);
+
+  // under construction...
+  
+  PG_RETURN_POINTER(stats);
+}
+
 Datum
 pgs2beginscan(PG_FUNCTION_ARGS)
 {
@@ -451,6 +582,10 @@ pgs2gettuple(PG_FUNCTION_ARGS)
   if (s->kill_prior_tuple) {
     rc = sen_index_del(ss->index->index, &(s->currentItemData));
     if (rc) { sen_log("sen_index_del failed(%d)", rc); }
+    /*
+    rc = sen_sym_del(ss->index->index->keys, &(s->currentItemData));
+    if (rc) { sen_log("sen_sym_del failed(%d)", rc); }
+    */
   }
 
   res = sen_records_next(ss->records, &s->xs_ctup.t_self, KEY_SIZE, NULL);
@@ -482,8 +617,6 @@ pgs2getmulti(PG_FUNCTION_ARGS)
   PG_RETURN_BOOL(more);
 }
 
-#define MAX_N_SORT_RESULT 10000
-
 Datum
 pgs2rescan(PG_FUNCTION_ARGS)
 {
@@ -496,6 +629,10 @@ pgs2rescan(PG_FUNCTION_ARGS)
   index_info *ii;
   sen_encoding encoding;
   sen_sel_operator op;
+  const char *option;
+  int max_n_sort_result = 10000;
+
+  if (!key) { elog(ERROR, "pgsenna2: access method does not support scan without scankey."); }
 
   ItemPointerSetInvalid(&s->currentItemData);
   ItemPointerSetInvalid(&s->currentMarkData);
@@ -528,9 +665,10 @@ pgs2rescan(PG_FUNCTION_ARGS)
             s->keyData[i].sk_strategy,
             s->keyData[i].sk_subtype,
             s->keyData[i].sk_argument);
-    if(s->keyData[i].sk_argument != NULL) {
+    if(s->keyData[i].sk_argument != 0) {
       c = text2cstr(DatumGetTextP(s->keyData[i].sk_argument));
-      if ((q = sen_query_open(c, sen_sel_or, PGS2_MAX_N_EXPRS, encoding))) {
+      if ((q = sen_query_open(c, strlen(c), sen_sel_or,
+                              PGS2_MAX_N_EXPRS, encoding))) {
         rc = sen_query_exec(ii->index, q, r, op);
         sen_query_close(q);
       }
@@ -540,7 +678,16 @@ pgs2rescan(PG_FUNCTION_ARGS)
       elog(ERROR, "pgsenna2: sen_query_exec failed(%d)", rc);
     }
   }
-  sen_records_sort(r, MAX_N_SORT_RESULT, NULL);
+  option = GetConfigOption("ludia.max_n_sort_result");
+  if (option) {
+    max_n_sort_result = atoi(option);
+    if (max_n_sort_result <= 0) {
+      elog(ERROR, "pgsenna2: value of max_n_sort_result is invalid: %d", max_n_sort_result);
+    }
+  } else {
+      elog(WARNING, "pgsenna2: value of max_n_sort_result = %d", max_n_sort_result);
+  }
+  sen_records_sort(r, max_n_sort_result, NULL);
   s->opaque = scan_stat_open(r, ii);
   ii->s = s;
   PG_RETURN_VOID();
@@ -582,10 +729,18 @@ pgs2costestimate(PG_FUNCTION_ARGS)
   PlannerInfo *root = (PlannerInfo *) PG_GETARG_POINTER(0);
   IndexOptInfo *index = (IndexOptInfo *) PG_GETARG_POINTER(1);
   List *indexQuals = (List *) PG_GETARG_POINTER(2);
+#ifdef POSTGRES82
+  //  RelOptInfo *outer_rel = (RelOptInfo *) PG_GETARG_POINTER(3);
+  Cost *indexStartupCost = (Cost *) PG_GETARG_POINTER(4);
+  Cost *indexTotalCost = (Cost *) PG_GETARG_POINTER(5);
+  Selectivity *indexSelectivity = (Selectivity *) PG_GETARG_POINTER(6);
+  double *indexCorrelation = (double *) PG_GETARG_POINTER(7);
+#else
   Cost *indexStartupCost = (Cost *) PG_GETARG_POINTER(3);
   Cost *indexTotalCost = (Cost *) PG_GETARG_POINTER(4);
   Selectivity *indexSelectivity = (Selectivity *) PG_GETARG_POINTER(5);
   double *indexCorrelation = (double *) PG_GETARG_POINTER(6);
+#endif
   double numIndexTuples = 0.0;
   double numIndexPages = 0.0;
   QualCost index_qual_cost;
@@ -631,22 +786,130 @@ pgs2costestimate(PG_FUNCTION_ARGS)
   *indexStartupCost = qual_arg_cost;
   *indexTotalCost += qual_arg_cost;
   *indexTotalCost += numIndexTuples * (cpu_index_tuple_cost + qual_op_cost);
+#ifdef POSTGRES82
+  *indexTotalCost += -DEFAULT_RANDOM_PAGE_COST;
+#endif
   *indexCorrelation = 1.0;
   sen_log("cost=(%f,%f,%f)", *indexStartupCost, *indexTotalCost, *indexSelectivity);
   PG_RETURN_VOID();
 }
 
+typedef struct _cons {
+  int8_t type ;
+  int8_t op;
+  int8_t weight;
+  int8_t mode;
+  int option;
+  struct _cons *cdr;
+  union {
+    struct {
+      const char *start;
+      unsigned int len;
+    } token;
+    struct {
+      struct _cons *car;
+      struct _cons **tail;
+    } expr;
+  } u;
+} cons;
+
+struct _sen_query {
+  const char *str;
+  const char *cur;
+  const char *str_end;
+  char *buffer;
+  unsigned int max_size;
+  sen_sel_operator default_op;
+  sen_sel_mode default_mode;
+  int escalation_threshold;
+  int escalation_decaystep;
+  int weight_offset;
+  sen_encoding encoding;
+  cons *expr;
+  int max_exprs;
+  int cur_expr;
+  cons cons_pool[1]; /* dummy */
+};
+
+static bool exec_query_noindex(sen_query *q, char *ct)
+{
+  bool res = false;
+  cons *token = q->expr->u.expr.car;
+  if (token) {
+    sen_sel_operator t_op;
+    int len;
+    int max_len = 16;
+    char *cquery = palloc(sizeof(char) * max_len);
+    for (t_op = sen_sel_or;; t_op = token->op) {
+      len = (token->u.token.len + 1) * sizeof(char);
+      if (len > max_len) {
+        max_len = len;
+        cquery = repalloc(cquery, max_len);
+      }
+      memcpy(cquery, token->u.token.start, len);
+      cquery[token->u.token.len] = '\0';
+      switch (t_op) {
+      case sen_sel_or :
+        res = res || strstr(ct, cquery);
+        break;
+      case sen_sel_and :
+        res = res && strstr(ct, cquery);
+        break;
+      case sen_sel_but :
+        res = res && !strstr(ct, cquery);
+        break;
+      default :
+        break;
+      }
+      if (!(token = token->cdr)) {
+        break;
+      }
+    }
+  }
+  return res;
+}
+
 Datum 
 pgs2contain(PG_FUNCTION_ARGS)
 {
   text *t = PG_GETARG_TEXT_P(0);
   text *q = PG_GETARG_TEXT_P(1);
-  char *ct = text2cstr(t);
-  char *cq = text2cstr(q);
-  bool res = strstr(ct, cq) ? true : false;
-  pfree(ct);
-  pfree(cq);
-  PG_RETURN_BOOL(res);
+  const char *option;
+
+  option = GetConfigOption("ludia.enable_seqscan");
+  if (option) {
+    if (!strcmp(option, "on")) {
+      char *ct = text2cstr(t);
+      char *cq = text2cstr(q);
+      bool res = false;
+      sen_encoding encoding;
+      sen_query *sq;
+      switch (GetDatabaseEncoding()) {
+      case PG_UTF8:
+        encoding = sen_enc_utf8;
+        break;
+      case PG_EUC_JP:
+        encoding = sen_enc_euc_jp;
+        break;
+      case PG_SJIS:
+        encoding = sen_enc_sjis;
+        break;
+      default:
+        encoding = sen_enc_default;
+      }
+      if ((sq = sen_query_open(cq, strlen(cq), sen_sel_or,
+                               PGS2_MAX_N_EXPRS, encoding))) {
+        res = exec_query_noindex(sq, ct);
+        sen_query_close(sq);
+      }
+      pfree(ct);
+      pfree(cq);
+      PG_RETURN_BOOL(res);
+    }
+    ereport(ERROR, (errmsg("pgsenna2: sequencial scan disabled.")));
+  }
+  ereport(ERROR, (errmsg("pgsenna2: value of enable_seqscan is invalid.")));
+  PG_RETURN_BOOL(false);
 }
 
 Datum 
@@ -664,13 +927,14 @@ pgs2getscore(PG_FUNCTION_ARGS)
   ItemPointer tid = (ItemPointer)PG_GETARG_POINTER(0);
   text *t = PG_GETARG_TEXT_P(1);
   char *indexname = text2cstr(t);
-  for (i = PGS2_MAX_N_INDEX_CACHE, ii = index_cache; i; i--, ii++) {
+  for (i = max_n_index_cache, ii = index_cache; i; i--, ii++) {
     if (ii->index && !strcmp(ii->name, indexname)) { break; }
   }
   if (!i || !ii->s || !(ss = ii->s->opaque) || !ss->records) {
     elog(ERROR, "pgsenna2: index (%s) is not available", indexname);
   }
   result = sen_records_find(ss->records, (void *)tid);
+  pfree(indexname);
   PG_RETURN_INT32(result);
 }
 
@@ -691,61 +955,101 @@ pgs2getnhits(PG_FUNCTION_ARGS)
   PG_RETURN_INT32(last_nhits);
 }
 
-static void
-do_dir(const char *dirname, int *n)
+#ifdef POSTGRES82
+Datum
+pgs2options(PG_FUNCTION_ARGS)
 {
-  DIR *dir = opendir(dirname);
-  if (dir) {
-    int fd = dirfd(dir);
-    struct dirent *ent;
-    fchdir(fd);
-    while ((ent = readdir(dir))) {
-      const char *name = ent->d_name;
-      if (!strcmp(name, ".") || !strcmp(name, "..")) {
-        continue;
-      }
-      switch (ent->d_type) {
-      case DT_REG :
-        {
-          int len = strlen(name) - 4;
-          if (len > 0 &&
-              name[len] == '.' &&
-              name[len + 1] == 'S' &&
-              name[len + 2] == 'E' &&
-              name[len + 3] == 'N') {
-            struct stat s;
-            char *idxname = strdup(name);
-            idxname[len] = '\0';
-            if (stat(idxname, &s) == -1 && errno == ENOENT) {
-              sen_index_remove(idxname);
-              (*n)++;
-            }
-          }
-        }
-        break;
-      case DT_DIR :
-      case DT_LNK :
-        do_dir(ent->d_name, n);
-        fchdir(fd);
-        break;
-      default :
+  elog(ERROR, "pgsenna2: pgs2options is called");
+  /*
+  Datum      reloptions = PG_GETARG_DATUM(0);
+  bool       validate = PG_GETARG_BOOL(1);
+  bytea      *result;
+
+  result = default_reloptions(reloptions, validate,
+                              PGS2_MIN_FILLFACTOR,
+                              PGS2_DEFAULT_FILLFACTOR);
+  if (result)
+    PG_RETURN_BYTEA_P(result);
+  */
+  PG_RETURN_NULL();
+}
+#endif
+
+Datum
+pgs2indexcache(PG_FUNCTION_ARGS)
+{
+  int n = 0;
+
+  FuncCallContext     *funcctx;
+  int                  call_cntr;
+  int                  max_calls;
+  TupleDesc            tupdesc;
+  AttInMetadata       *attinmeta;
+  index_info **last_used_cache_array;
+  
+  if (SRF_IS_FIRSTCALL()) {
+    index_info *tmp_used_cache;
+    MemoryContext   oldcontext;
+    funcctx = SRF_FIRSTCALL_INIT();
+    oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);
+    if (get_call_result_type(fcinfo, NULL, &tupdesc) != TYPEFUNC_COMPOSITE)
+      ereport(ERROR,
+              (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+               errmsg("function returning record called in context "
+                      "that cannot accept type record")));
+    attinmeta = TupleDescGetAttInMetadata(tupdesc);
+    funcctx->attinmeta = attinmeta;
+
+    last_used_cache_array = (index_info **)palloc(max_n_index_cache * sizeof(index_info *));
+    tmp_used_cache = last_used_cache;
+    while(tmp_used_cache) {
+      last_used_cache_array[n] = tmp_used_cache;
+      tmp_used_cache = tmp_used_cache->next;
+      n++;
+      if(n > max_n_index_cache) {
+        elog(WARNING, "pgsenna2: warning happened at index cache");
         break;
       }
     }
-    closedir(dir);
+    funcctx->user_fctx = (void *)last_used_cache_array;
+    funcctx->max_calls = n;
+    MemoryContextSwitchTo(oldcontext);
   }
-}
 
-Datum
-pgs2destroy(PG_FUNCTION_ARGS)
-{
-  int n = 0;
-  DIR *dir = opendir(".");
-  if (dir) {
-    int fd = dirfd(dir);
-    do_dir(DataDir, &n);
-    fchdir(fd);
-    closedir(dir);
+  funcctx = SRF_PERCALL_SETUP();
+
+  last_used_cache_array = (index_info **) funcctx->user_fctx;
+  
+  call_cntr = funcctx->call_cntr;
+  max_calls = funcctx->max_calls;
+  attinmeta = funcctx->attinmeta;
+ 
+  if (call_cntr < max_calls) {
+    int i;
+    char       **values;
+    HeapTuple    tuple;
+    Datum        result;
+    
+    values = (char **) palloc(12 * sizeof(char *));
+    for (i = 0; i < 6; i++) {
+      values[i] = (char *) palloc(16 * sizeof(char));
+    }
+    last_used_cache_array = (index_info **) funcctx->user_fctx;
+    snprintf(values[0], 16, "%d", last_used_cache_array[call_cntr]->relid);
+    snprintf(values[1], 16, "%d", last_used_cache_array[call_cntr]->namespace);
+    snprintf(values[2], 16, "%s", last_used_cache_array[call_cntr]->name);
+    snprintf(values[3], 16, "%d", (int)last_used_cache_array[call_cntr]->previous);
+    snprintf(values[4], 16, "%d", (int)last_used_cache_array[call_cntr]);
+    snprintf(values[5], 16, "%d", (int)last_used_cache_array[call_cntr]->next);
+    tuple = BuildTupleFromCStrings(attinmeta, values);
+    result = HeapTupleGetDatum(tuple);
+    for (i = 0; i < 6; i++) {
+      pfree(values[i]);
+    }
+    pfree(values);
+
+    SRF_RETURN_NEXT(funcctx, result);
+  } else {
+    SRF_RETURN_DONE(funcctx);
   }
-  PG_RETURN_INT32(n);
 }
